'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest
from io import StringIO
from contextlib import redirect_stdout

import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer

from dmba import printTermDocumentMatrix



class TestTextMining(unittest.TestCase):
  def test_printTermDocumentMatrix(self):
    text = ['this is the first sentence.',
            'this is a second sentence.',
            'the third sentence is here.']
    count_vect = CountVectorizer()
    counts = count_vect.fit_transform(text)

    pd.set_option('display.width', 100)
    pd.set_option('display.max_columns', 20)
    
    out = StringIO()
    with redirect_stdout(out):
      printTermDocumentMatrix(count_vect, counts)
    s = out.getvalue()
    self.assertIn('S1  S2  S3', s)
    self.assertIn('first      1   0   0', s)
    self.assertIn('the        1   0   1', s)



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck
'''
import math
import numpy as np
from sklearn.metrics import classification, regression


def adjusted_r2_score(y_true, y_pred, model):
    """ calculate adjusted R2
    Input:
        y_true: actual values
        y_pred: predicted values
        model: predictive model
    """
    n = len(y_pred)
    p = len(model.coef_)
    if p >= n - 1:
      return 0
    r2 = regression.r2_score(y_true, y_pred)
    return 1 - (1 - r2) * (n - 1) /(n - p - 1)
    
def AIC_score(y_true, y_pred, model=None, df=None):
    """ calculate Akaike Information Criterion (AIC) 
    Input:
        y_true: actual values
        y_pred: predicted values
        model (optional): predictive model
        df (optional): degrees of freedom of model
        
    One of model or df is requried
    """
    if df is None and model is None:
      raise ValueError('You need to provide either model or df')
    n = len(y_pred)
    p = len(model.coef_) + 1 if df is None else df
    resid = np.array(y_true) - np.array(y_pred)
    sse = np.sum(resid ** 2) 
    constant =  n + n * np.log(2*np.pi)
    return n * math.log(sse / n) + constant + 2 * (p + 1)


def BIC_score(y_true, y_pred, model=None, df=None):
    """ calculate Schwartz's Bayesian Information Criterion (AIC) 
    Input:
        y_true: actual values
        y_pred: predicted values
        model: predictive model
        df (optional): degrees of freedom of model
    """
    aic = AIC_score(y_true, y_pred, model=model, df=df)
    p = len(model.coef_) + 1 if df is None else df
    n = len(y_pred)
    return aic - 2 * (p + 1) + math.log(n) * (p + 1)  


def regressionSummary(y_true, y_pred):
    """ print regression performance metrics 
    
    Input:
        y_true: actual values
        y_pred: predicted values
    """
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    y_res = y_true - y_pred
    metrics = [
        ('Mean Error (ME)', sum(y_res) / len(y_res)),
        ('Root Mean Squared Error (RMSE)', math.sqrt(regression.mean_squared_error(y_true, y_pred))),
        ('Mean Absolute Error (MAE)', sum(abs(y_res)) / len(y_res)),
    ]
    if all(yt != 0 for yt in y_true):
      metrics.extend([
        ('Mean Percentage Error (MPE)', 100 * sum(y_res / y_true) / len(y_res)),
        ('Mean Absolute Percentage Error (MAPE)', 100 * sum(abs(y_res / y_true) / len(y_res))),
      ])
    fmt1 = '{{:>{}}} : {{:.4f}}'.format(max(len(m[0]) for m in metrics))
    print('\nRegression statistics\n')
    for metric, value in metrics:
        print(fmt1.format(metric, value))


def classificationSummary(y_true, y_pred, class_names=None):
    """ Print a summary of classification performance
    
    Input:
        y_true: actual values
        y_pred: predicted values
        class_names (optional): list of class names
    """
    confusionMatrix = classification.confusion_matrix(y_true, y_pred)
    accuracy = classification.accuracy_score(y_true, y_pred)

    print('Confusion Matrix (Accuracy {:.4f})\n'.format(accuracy))
    
    # Pretty-print confusion matrix
    cm = confusionMatrix

    labels = class_names
    if labels is None:
        labels = [str(i) for i in range(len(cm))]
   
    # Convert the confusion matrix and labels to strings
    cm = [[str(i) for i in row] for row in cm]
    labels = [str(i) for i in labels]

    # Determine the width for the first label column and the individual cells    
    prediction = 'Prediction'
    actual = 'Actual'
    labelWidth = max(len(s) for s in labels)
    cmWidth = max(max(len(s) for row in cm for s in row), labelWidth) + 1
    labelWidth = max(labelWidth, len(actual))
    
    # Construct the format statements
    fmt1 = '{{:>{}}}'.format(labelWidth)
    fmt2 = '{{:>{}}}'.format(cmWidth) * len(labels)

    # And print the confusion matrix    
    print(fmt1.format(' ') + ' ' + prediction)
    print(fmt1.format(actual), end='')
    print(fmt2.format(*labels))
    
    for cls, row in zip(labels, cm):
        print(fmt1.format(cls), end='')
        print(fmt2.format(*row))



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest
from collections import namedtuple
from contextlib import redirect_stdout
from io import StringIO

from sklearn.metrics.regression import r2_score

from dmba import adjusted_r2_score, AIC_score, BIC_score
from dmba import regressionSummary, classificationSummary


MockModel = namedtuple('Model', 'coef_')


class TestMetric(unittest.TestCase):
  def test_adjusted_r2_score(self):
    y_true = [1, 2, 3, 4, 5]
    y_pred = [1, 3, 2, 5, 4]

    r2 = r2_score(y_true, y_pred)
    assert r2 == 0.6

    for df in range(4):
      n = len(y_true)
      f = (n - 1) / (n - df - 1)
      expected = 1 - (1 - r2) * f
      self.assertAlmostEqual(
          adjusted_r2_score(y_true, y_pred, MockModel(coef_=[1] * df)),
          expected, places=3, msg=f'failed for df={df}')

    # if degree of freedom gets too large returns 0
    coef = [1] * 4
    self.assertAlmostEqual(
        adjusted_r2_score(y_true, y_pred, MockModel(coef_=coef)),
        0, places=3, msg=f'failed for df={df}')

    coef = [1] * 5
    self.assertAlmostEqual(
        adjusted_r2_score(y_true, y_pred, MockModel(coef_=coef)),
        0, places=3, msg=f'failed for df={df}')

  def test_AIC_score(self):
    y_true = [1, 2, 3, 4, 5]
    y_pred = [1, 3, 2, 5, 4]

    self.assertAlmostEqual(
        AIC_score(y_true, y_pred, MockModel(coef_=[1] * 2)),
        21.0736, places=3)

    self.assertAlmostEqual(
        AIC_score(y_true, y_pred, df=3),
        AIC_score(y_true, y_pred, MockModel(coef_=[1] * 2)), places=3)

    self.assertGreater(
        AIC_score(y_true, y_pred, df=3),
        AIC_score(y_true, y_pred, df=2))

  def test_BIC_score(self):
    y_true = [1, 2, 3, 4, 5]
    y_pred = [1, 3, 2, 5, 4]

    self.assertAlmostEqual(
        BIC_score(y_true, y_pred, MockModel(coef_=[1] * 2)),
        19.51141, places=3)

    self.assertAlmostEqual(
        BIC_score(y_true, y_pred, df=3),
        BIC_score(y_true, y_pred, MockModel(coef_=[1] * 2)), places=3)

    self.assertGreater(
        BIC_score(y_true, y_pred, df=3),
        BIC_score(y_true, y_pred, df=2))

  def test_regressionSummary(self):
    y_true = [1, 2, 3, 4, 5]
    y_pred = [1, 3, 2, 5, 4]

    out = StringIO()
    with redirect_stdout(out):
      regressionSummary(y_true, y_pred)
    s = out.getvalue()
    self.assertIn('Regression statistics', s)
    self.assertIn('(ME) : 0.0000', s)
    self.assertIn('(RMSE) : 0.8944', s)
    self.assertIn('(MAE) : 0.8000', s)
    self.assertIn('(MPE) : -4.3333', s)
    self.assertIn('(MAPE) : 25.6667', s)

    y_true = [0, 1, 2, 3, 4]
    y_pred = [0, 2, 1, 4, 3]

    out = StringIO()
    with redirect_stdout(out):
      regressionSummary(y_true, y_pred)
    s = out.getvalue()
    self.assertIn('Regression statistics', s)
    self.assertIn('(ME) : 0.0000', s)
    self.assertIn('(RMSE) : 0.8944', s)
    self.assertIn('(MAE) : 0.8000', s)
    self.assertNotIn('(MPE)', s)
    self.assertNotIn('(MAPE)', s)

  def test_classificationSummary(self):
    y_true = [1, 0, 0, 1, 1, 1]
    y_pred = [1, 0, 1, 1, 0, 0]

    out = StringIO()
    with redirect_stdout(out):
      classificationSummary(y_true, y_pred, class_names=['a', 'b'])
    s = out.getvalue()

    self.assertIn('Confusion Matrix', s)
    self.assertIn('       Prediction', s)
    self.assertIn('a 1 1', s)
    self.assertIn('b 2 2', s)
